@@ -697,6 +697,15 @@ class RegExpStringIteratorObject : public NativeObject {
 RegExpStringIteratorObject* NewRegExpStringIteratorTemplate(JSContext* cx);
 RegExpStringIteratorObject* NewRegExpStringIterator(JSContext* cx);
 
+#ifdef NIGHTLY_BUILD
+class IteratorRangeObject : public NativeObject {
+ public:
+  static const JSClass class_;
+};
+
+IteratorRangeObject* NewIteratorRange(JSContext* cx);
+#endif
+
 [[nodiscard]] bool EnumerateProperties(JSContext* cx, HandleObject obj,
                                        MutableHandleIdVector props);
 
@@ -751,6 +760,8 @@ class IteratorObject : public NativeObject {
  public:
   static const JSClass class_;
   static const JSClass protoClass_;
+
+  static bool finishInit(JSContext* cx, HandleObject ctor, HandleObject proto);
 };
 
 /*
@@ -787,12 +798,23 @@ class IteratorHelperObject : public NativeObject {
     // Never null.
     GeneratorSlot,
 
+    // The [[UnderlyingIterator]] internal slot. Either an object or undefined
+    // in the case of IteratorConcat. In the spec, the internal slot stores an
+    // "Iterator Record", but our implementation only stores the actual iterator
+    // object.
+    UnderlyingIteratorSlot,
+
     SlotCount,
   };
 
   static_assert(GeneratorSlot == ITERATOR_HELPER_GENERATOR_SLOT,
                 "GeneratorSlot must match self-hosting define for generator "
                 "object slot.");
+
+  static_assert(UnderlyingIteratorSlot ==
+                    ITERATOR_HELPER_UNDERLYING_ITERATOR_SLOT,
+                "UnderlyingIteratorSlot must match self-hosting define for "
+                "underlying iterator slot.");
 };
 
 IteratorHelperObject* NewIteratorHelper(JSContext* cx);
@@ -800,6 +822,16 @@ IteratorHelperObject* NewIteratorHelper(JSContext* cx);
 bool IterableToArray(JSContext* cx, HandleValue iterable,
                      MutableHandle<ArrayObject*> array);
 
+bool HasOptimizableArrayIteratorPrototype(JSContext* cx);
+
+enum class MustBePacked { No, Yes };
+
+template <MustBePacked Packed>
+bool IsArrayWithDefaultIterator(JSObject* obj, JSContext* cx);
+
+bool IsMapObjectWithDefaultIterator(JSObject* obj, JSContext* cx);
+bool IsSetObjectWithDefaultIterator(JSObject* obj, JSContext* cx);
+
 // Typed arrays and classes with an enumerate hook can have extra properties not
 // included in the shape's property map or the object's dense elements.
 static inline bool ClassCanHaveExtraEnumeratedProperties(const JSClass* clasp) {
